/*
Module : SCardWrappers.h
Purpose: Defines the interface for a set of C++ class which encapsulate the Windows Smart Card APIs.
History: PJN / 09-03-2013 1. Initial creation
         PJN / 05-04-2015 1. Updated copyright details
                          2. Updated the code to compile cleanly on VC 2013
                          3. Added SAL annotations to all the code
         PJN / 29-11-2015 1. Updated the code to compile cleanly on VC 2015
                          2. Verified there were no changes to the Windows Smart Card API on Windows 10.
                          3. Added SAL annotations to the ListReaderGroups methods for the return value
         PJN / 02-01-2016 1. Updated copyright details.
                          2. The code now uses wostringstream and ostringstream instead of basic_stringstream<wchar_t>
                          and basic_stringstream<char>.
         PJN / 24-07-2016 1. Reworked the AttributeASCIIBinaryStringToString function to avoid the need to call the 
                          Win32 API MultiByteToWideChar. Instead now the code uses the ATL class CA2T
         PJN / 25-03-2017 1. Updated copyright details.
                          2. Updated code to compile cleanly on VC 2017
         PJN / 12-11-2017 1. Renamed the CSCARDWRAPPERS_MFC_EXTENSIONS define to "SCARDWRAPPERS_MFC_EXTENSIONS" to 
                          be consistent with the author's other code.
                          2. Replaced CString::operator LPC*STR() calls with CString::GetString calls
                          3. Replaced NULL with nullptr throughout the codebase. This means that the code now
                          requires VC 2010 at a minimum to compile.
                          4. Replaced BOOL with bool throughout the codebase.
                          5. Removed unnecessary AccessStartedEvent & ReleaseStartedEvent methods from CContext
                          6. Updated parameters to CContext::GetProviderName & CContext2::GetProviderName to match the SDK
                          7. Removed redundant CCard & CCard2 methods which call CContext::GetAttrib
                          8. Reviewed all the SAL annotations to make sure they are in sync with the latest Windows 10 SDK
                          9. Reviewed the code and made sure it supports everything in the latest Windows 10 SDK
                          10. Moved IntToHex, HexToInt, BinaryToString, AttributeBinaryToString, 
                          AttributeBinaryStringToString, AttributeASCIIBinaryStringToString & StringToBinary functions 
                          from SCardWrappers.h to the test.cpp demo app module.
         PJN / 11-07-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
                          3. Removed code which supported SCARDWRAPPERS_MFC_EXTENSIONS code path
         PJN / 16-09-2018 1. Fixed a number of compiler warnings when using VS 2017 15.8.4
         PJN / 21-04-2019 1. Updated copyright details
                          2. Updated the code to clean compile on VC 2019
         PJN / 10-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 15-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 12-04-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 16-08-2021 1. Updated copyright details.
         PJN / 06-03-2022 1. Updated copyright details
                          2. Updated the code to use C++ uniform initialization for all variable declarations.
                          3. Removed all code which used LoadLibrary and GetProcAddress.
         PJN / 18-12-2023 1. Updated copyright details.
                          2. Remove usage of __if_exists from module.

Copyright (c) 2013 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////////////// Macros / Includes ////////////////////////////////

#ifndef __SCARDWRAPPERS_H__
#define __SCARDWRAPPERS_H__

#ifndef CSCARDWRAPPERS_EXT_CLASS
#define CSCARDWRAPPERS_EXT_CLASS
#endif //#ifndef CSCARDWRAPPERS_EXT_CLASS

#ifndef CSCARDWRAPPERS_EXT_API
#define CSCARDWRAPPERS_EXT_API
#endif //#ifndef CSCARDWRAPPERS_EXT_API

#pragma once

#ifndef _WINSCARD_H_
#pragma message("To avoid this message, please put winscard.h in your pre compiled header (normally stdafx.h)")
#include <winscard.h>
#endif //#ifndef _WINSCARD_H_

#ifndef _EXCEPTION_
#pragma message("To avoid this message, please put exception in your pre compiled header (normally stdafx.h)")
#include <exception>
#endif //#ifndef _EXCEPTION_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef __ATLBASE_H__
#pragma message("To avoid this message, please put atlbase.h in your pre compiled header (normally stdafx.h)")
#include <atlbase.h>
#endif //#ifndef __ATLBASE_H__

#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#include <atlstr.h>
#endif //#ifndef __ATLSTR_H__

#pragma comment(lib, "winscard.lib")


/////////////////////////// Classes ///////////////////////////////////////////

namespace SCard
{
//typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE
using StringArray = std::vector<String>;
using GUIDArray = std::vector<GUID>;
using ByteArray = std::vector<BYTE>;

inline void CSCARDWRAPPERS_EXT_API MultiSZToStringArray(_In_opt_ LPCTSTR pszString, _Inout_ StringArray& output)
{
  //Clear the vector
  output.clear();
  if (pszString == nullptr)
    return;

  //Add each element we find
#pragma warning(suppress: 26429)
  LPCTSTR pszValue{pszString};
#pragma warning(suppress: 26489)
  while (*pszValue != _T('\0'))
  {
#pragma warning(suppress: 26486 26489)
    output.push_back(pszValue);
#pragma warning(suppress: 26481 26486)
    pszValue += (_tcslen(pszValue) + 1);
  }
}


//Class which encapsulates an SCard exception
class CSCARDWRAPPERS_EXT_CLASS CException : public std::exception
{
public:
//Constructors / Destructors
  CException(_In_ LONG nValue) noexcept : m_nValue{nValue}
  {
  }
  CException(const CException&) = default;
  CException(CException&&) = default;
  ~CException() = default; //NOLINT(modernize-use-override)

//Methods
  CException& operator=(const CException&) = default;
  CException& operator=(CException&&) = default;

//Members variables
  LONG m_nValue;
#pragma warning(suppress: 26495)
};


inline void CSCARDWRAPPERS_EXT_API CheckException(_In_ LONG nRet)
{
  if (nRet != SCARD_S_SUCCESS)
  {
    TRACE(_T("Warning: throwing SCard::CException for error %d\n"), nRet);
    throw SCard::CException(nRet);
  }
}


//Class which encapsulates a SCard context handle
class CSCARDWRAPPERS_EXT_CLASS CContext
{
public:
//Constructors / Destructors
#pragma warning(suppress: 26455)
  CContext() noexcept : m_ctx{NULL}
  {
  }

  CContext(_In_ const CContext& ctx) = delete;

  CContext(_In_ CContext&& ctx) noexcept : CContext()
  {
    Attach(ctx.Detach());
  }

  explicit CContext(_In_ SCARDCONTEXT ctx) noexcept : CContext()
  {
    m_ctx = ctx;
  }

  ~CContext() noexcept
  {
    if (m_ctx != NULL)
      Close();
  }

//Methods
  CContext& operator=(_In_ const CContext& ctx) = delete;

  CContext& operator=(_In_ CContext&& ctx) noexcept
  {
    if (m_ctx != NULL)
      Close();
    Attach(ctx.Detach());
    return *this;
  }

  LONG Create(_In_ DWORD dwScope = SCARD_SCOPE_USER) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx == NULL);

    return SCardEstablishContext(dwScope, nullptr, nullptr, &m_ctx);
  }

  LONG Close() noexcept
  {
    LONG nRet{SCARD_S_SUCCESS};
    if (m_ctx != NULL)
    {
      nRet = SCardReleaseContext(m_ctx);
      m_ctx = 0;
    }
    return nRet;
  }

  LONG IsValid() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardIsValidContext(m_ctx);
  }

  operator SCARDCONTEXT() const noexcept
  {
    return m_ctx;
  }

  void Attach(_In_ SCARDCONTEXT ctx) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx == NULL);

    m_ctx = ctx;
  }

  SCARDCONTEXT Detach() noexcept
  {
    const SCARDCONTEXT ctx{m_ctx};
    m_ctx = NULL;
    return ctx;
  }

  LONG AddReaderToGroup(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szGroupName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardAddReaderToGroup(m_ctx, szReaderName, szGroupName);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG Audit(_In_ DWORD dwEvent) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardAudit(m_ctx, dwEvent);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

  LONG Cancel() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardCancel(m_ctx);
  }

  LONG ForgetCardType(_In_ LPCTSTR szCardName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardForgetCardType(m_ctx, szCardName);
  }

  LONG ForgetReader(_In_ LPCTSTR szReaderName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardForgetReader(m_ctx, szReaderName);
  }

  LONG ForgetReaderGroup(_In_ LPCTSTR szGroupName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardForgetReaderGroup(m_ctx, szGroupName);
  }

  LONG FreeMemory(_In_ LPCVOID pvMem) noexcept
  {
    return SCardFreeMemory(m_ctx, pvMem);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG GetProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId,
                                                            _When_(_Old_(*pcchProvider) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)szProvider, _Outptr_result_buffer_all_(*pcchProvider)))
                                                            _When_(_Old_(*pcchProvider) != SCARD_AUTOALLOCATE, _Out_writes_to_(*pcchProvider, *pcchProvider) _Post_z_)
                                                            LPTSTR szProvider, _Inout_ LPDWORD pcchProvider) noexcept
  {
    return SCardGetCardTypeProviderName(m_ctx, szCardName, dwProviderId, szProvider, pcchProvider);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG GetProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId, _Out_ String& sProviderName)
  {
    //First get the length of the data
    DWORD dwProvider{SCARD_AUTOALLOCATE};
    LPTSTR szProvider{nullptr};
#pragma warning(suppress: 26490)
    const LONG nRet{GetProviderName(szCardName, dwProviderId, reinterpret_cast<LPTSTR>(&szProvider), &dwProvider)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    sProviderName = szProvider;
    FreeMemory(szProvider);

    return nRet;
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG GetDeviceTypeId(_In_ LPCTSTR szReaderName, _Inout_ LPDWORD pdwDeviceTypeId) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardGetDeviceTypeId(m_ctx, szReaderName, pdwDeviceTypeId);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

  LONG GetProviderId(_In_ LPCTSTR szCard, _Out_ LPGUID pguidProviderId) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardGetProviderId(m_ctx, szCard, pguidProviderId);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG GetReaderDeviceInstanceId(_In_ LPCTSTR szReaderName,
                                                                      _When_(_Old_(*pcchDeviceInstanceId) == SCARD_AUTOALLOCATE, _At_((LPTSTR *)szDeviceInstanceId, _Outptr_result_buffer_maybenull_(*pcchDeviceInstanceId) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                                                      _When_(_Old_(*pcchDeviceInstanceId) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchDeviceInstanceId) _Post_ _NullNull_terminated_)
                                                                      LPTSTR szDeviceInstanceId, _Inout_ LPDWORD pcchDeviceInstanceId) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

#pragma warning(suppress: 6387)
    return SCardGetReaderDeviceInstanceId(m_ctx, szReaderName, szDeviceInstanceId, pcchDeviceInstanceId);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG GetReaderDeviceInstanceId(_In_ LPCTSTR szReaderName, _Out_ String& sDeviceInstanceId)
  {
    //First get the length of the data
    DWORD dwDeviceInstanceID{SCARD_AUTOALLOCATE};
    LPTSTR szDeviceInstanceId{nullptr};
#pragma warning(suppress: 26490)
    const LONG nRet{GetReaderDeviceInstanceId(szReaderName, reinterpret_cast<LPTSTR>(&szDeviceInstanceId), &dwDeviceInstanceID)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
#pragma warning(suppress: 26477)
    ATLASSUME(szDeviceInstanceId != nullptr);
    sDeviceInstanceId = szDeviceInstanceId;
    if (szDeviceInstanceId != nullptr)
      FreeMemory(szDeviceInstanceId);

    return nRet;
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG GetReaderIcon(_In_ LPCTSTR szReaderName,
                                                          _When_(_Old_(*pcbIcon) == SCARD_AUTOALLOCATE, _At_((LPBYTE*)pbIcon, _Outptr_result_bytebuffer_all_maybenull_(*pcbIcon)))
                                                          _When_(_Old_(*pcbIcon) != SCARD_AUTOALLOCATE, _Out_writes_bytes_to_(*pcbIcon, *pcbIcon) _Post_z_)
                                                          LPBYTE pbIcon, _Inout_ LPDWORD pcbIcon) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardGetReaderIcon(m_ctx, szReaderName, pbIcon, pcbIcon);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG GetReaderIcon(_In_ LPCTSTR szReaderName, _Out_ ByteArray& icon)
  {
    //First get the length of the data
    DWORD dwIcon{SCARD_AUTOALLOCATE};
    LPBYTE pbIcon{nullptr};
#pragma warning(suppress: 26490)
    const LONG nRet{GetReaderIcon(szReaderName, reinterpret_cast<LPBYTE>(&pbIcon), &dwIcon)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;

    icon.clear();
#pragma warning(suppress: 26444 26481 26489)
    icon.insert(icon.end(), pbIcon, pbIcon + dwIcon);
    if (pbIcon != nullptr)
      FreeMemory(pbIcon);

    return nRet;
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

  LONG GetStatusChange(_In_ DWORD dwTimeout, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardGetStatusChange(m_ctx, dwTimeout, rgReaderStates, cReaders);
  }

  LONG IntroduceCardType(_In_ LPCTSTR szCardName, _In_opt_  LPCGUID pguidPrimaryProvider, _In_opt_ LPCGUID rgguidInterfaces,
                         _In_ DWORD dwInterfaceCount, _In_ LPCBYTE pbAtr, _In_ LPCBYTE pbAtrMask, _In_ DWORD cbAtrLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardIntroduceCardType(m_ctx, szCardName, pguidPrimaryProvider, rgguidInterfaces, dwInterfaceCount, pbAtr, pbAtrMask, cbAtrLen);
  }

  LONG IntroduceReader(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szDeviceName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardIntroduceReader(m_ctx, szReaderName, szDeviceName);
  }

  LONG IntroduceReaderGroup(_In_ LPCTSTR szGroupName) noexcept
  {
    return SCardIntroduceReaderGroup(m_ctx, szGroupName);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListCards(_In_opt_ LPCBYTE pbAtr, _In_reads_opt_(cguidInterfaceCount) LPCGUID rgguidInterfaces, _In_ DWORD cguidInterfaceCount,
                                                      _When_(_Old_(*pcchCards) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszCards, _Outptr_result_buffer_maybenull_(*pcchCards) _At_(*_Curr_, _Post_ _NullNull_terminated_)))
                                                      _When_(_Old_(*pcchCards) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchCards) _Post_ _NullNull_terminated_) LPTSTR mszCards, _Inout_ LPDWORD pcchCards) noexcept
  {
  #pragma warning(suppress: 6387)
    return SCardListCards(m_ctx, pbAtr, rgguidInterfaces, cguidInterfaceCount, mszCards, pcchCards);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListCards(_In_opt_ LPCBYTE pbAtr, _In_reads_opt_(cguidInterfaceCount) LPCGUID rgguidInterfaces, _In_ DWORD cguidInterfaceCount, _Out_ StringArray& cards)
  {
    LPTSTR mszCards{nullptr};
    DWORD dwCards{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{ListCards(pbAtr, rgguidInterfaces, cguidInterfaceCount, reinterpret_cast<LPTSTR>(&mszCards), &dwCards)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    MultiSZToStringArray(mszCards, cards);
    if (mszCards != nullptr)
      FreeMemory(mszCards);

    return nRet;
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListInterfaces(_In_ LPCTSTR szCard, _Out_ LPGUID pguidInterfaces, _Inout_ LPDWORD pcguidInterfaces) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardListInterfaces(m_ctx, szCard, pguidInterfaces, pcguidInterfaces);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListInterfaces(_In_ LPCTSTR szCard, _Out_ GUIDArray& interfaces)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

#pragma warning(suppress: 26429)
    LPGUID pGuidInterfaces{nullptr};
    DWORD dwGuidInterfaces{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{ListInterfaces(szCard, reinterpret_cast<LPGUID>(&pGuidInterfaces), &dwGuidInterfaces)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;

    interfaces.reserve(dwGuidInterfaces);
    for (DWORD i=0; i<dwGuidInterfaces; i++)
#pragma warning(suppress: 26481 26489)
      interfaces.push_back(pGuidInterfaces[i]);
    FreeMemory(pGuidInterfaces);

    return nRet;
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListReaderGroups(_Out_writes_opt_(*pcchGroups) _Post_ _NullNull_terminated_ LPTSTR mszGroups, _Inout_ LPDWORD pcchGroups) noexcept
  {
    return SCardListReaderGroups(m_ctx, mszGroups, pcchGroups);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListReaderGroups(_Out_ StringArray& groups)
  {
    LPTSTR mszGroups{nullptr};
    DWORD dwGroups{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{ListReaderGroups(reinterpret_cast<LPTSTR>(&mszGroups), &dwGroups)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    MultiSZToStringArray(mszGroups, groups);
    if (mszGroups != nullptr)
      FreeMemory(mszGroups);

    return nRet;
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListReaders(_In_opt_ LPCTSTR mszGroups,
                                                        _When_(_Old_(*pcchReaders) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaders, _Outptr_result_buffer_maybenull_(*pcchReaders) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                                        _When_(_Old_(*pcchReaders) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaders) _Post_ _NullNull_terminated_)
                                                        LPTSTR mszReaders, _Inout_ LPDWORD pcchReaders) noexcept
  {
  #pragma warning(suppress: 6387)
    return SCardListReaders(m_ctx, mszGroups, mszReaders, pcchReaders);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG ListReaders(_In_opt_ LPCTSTR mszGroups, _Out_ StringArray& readers)
  {
    LPTSTR mszReaders{nullptr};
    DWORD dwReaders{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{ListReaders(mszGroups, reinterpret_cast<LPTSTR>(&mszReaders), &dwReaders)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    MultiSZToStringArray(mszReaders, readers);
    if (mszReaders != nullptr)
      FreeMemory(mszReaders);

    return nRet;
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG ListReadersWithDeviceInstanceId(_In_ LPCTSTR szDeviceInstanceId, 
                                                                            _When_(_Old_(*pcchReaders) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaders, _Outptr_result_buffer_maybenull_(*pcchReaders) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                                                            _When_(_Old_(*pcchReaders) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaders) _Post_ _NullNull_terminated_)
                                                                            LPTSTR mszReaders, _Inout_ LPDWORD pcchReaders) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

#pragma warning(suppress: 6387)
    return SCardListReadersWithDeviceInstanceId(m_ctx, szDeviceInstanceId, mszReaders, pcchReaders);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
  _Success_(return == SCARD_S_SUCCESS) LONG ListReadersWithDeviceInstanceId(_In_ LPCTSTR szDeviceInstanceId, _Out_ StringArray& readers)
  {
    LPTSTR mszReaders{nullptr};
    DWORD dwReaders{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{ListReadersWithDeviceInstanceId(szDeviceInstanceId, reinterpret_cast<LPTSTR>(&mszReaders), &dwReaders)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    MultiSZToStringArray(mszReaders, readers);
    if (mszReaders != nullptr)
      FreeMemory(mszReaders);

    return nRet;
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

  LONG LocateCards(_In_ LPCTSTR mszCards, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardLocateCards(m_ctx, mszCards, rgReaderStates, cReaders);
  }

  LONG LocateCardsByATR(_In_ LPSCARD_ATRMASK rgAtrMasks, _In_ DWORD cAtrs, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardLocateCardsByATR(m_ctx, rgAtrMasks, cAtrs, rgReaderStates, cReaders);
  }

#if (NTDDI_VERSION >= NTDDI_VISTA)
#pragma warning(suppress: 26429)
  LONG ReadCache(_In_ UUID* pCardIdentifier, _In_ DWORD dwFreshnessCounter, _In_ LPTSTR szLookupName, _Out_writes_bytes_(*pdwDataLen) PBYTE pData, _Out_ DWORD* pdwDataLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

#pragma warning(suppress: 6001)
    return SCardReadCache(m_ctx, pCardIdentifier, dwFreshnessCounter, szLookupName, pData, pdwDataLen);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

#if (NTDDI_VERSION >= NTDDI_VISTA)
  LONG WriteCache(_In_ UUID* pCardIdentifier, _In_ DWORD dwFreshnessCounter, _In_ LPTSTR szLookupName, _In_reads_bytes_(dwDataLen) PBYTE pData, _In_ DWORD dwDataLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardWriteCache(m_ctx, pCardIdentifier, dwFreshnessCounter, szLookupName, pData, dwDataLen);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

  LONG RemoveReaderFromGroup(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szGroupName) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_ctx != NULL);

    return SCardRemoveReaderFromGroup(m_ctx, szReaderName, szGroupName);
  }

  LONG SetCardTypeProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId, _In_ LPCTSTR szProvider) noexcept
  {
    return SCardSetCardTypeProviderName(m_ctx, szCardName, dwProviderId, szProvider);
  }

//Member variables
  SCARDCONTEXT m_ctx;

protected:
  friend class CCard;
};


//exception based version of CContext
class CSCARDWRAPPERS_EXT_CLASS CContext2 : public CContext
{
public:
//Constructors / Destructors
  CContext2() noexcept //NOLINT(modernize-use-equals-default)
  {
  }
  CContext2(const CContext2&) = delete;
  CContext2(CContext2&&) = default;
  explicit CContext2(_In_ SCARDCONTEXT ctx) noexcept : CContext{ctx}
  {
  }
  ~CContext2() = default;

//Methods
  CContext2& operator=(const CContext2&) = delete;
#pragma warning(suppress: 26456)
  CContext2& operator=(CContext2&&) = default;

#pragma warning(suppress: 26434)
  void Create(_In_ DWORD dwScope = SCARD_SCOPE_USER)
  {
    const LONG nRet{__super::Create(dwScope)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void Close()
  {
    const LONG nRet{__super::Close()};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void AddReaderToGroup(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szGroupName)
  {
    const LONG nRet{__super::AddReaderToGroup(szReaderName, szGroupName)};
    CheckException(nRet);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void Audit(_In_ DWORD dwEvent)
  {
    const LONG nRet{__super::Audit(dwEvent)};
    CheckException(nRet);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#pragma warning(suppress: 26434)
  void Cancel()
  {
    const LONG nRet{__super::Cancel()};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void ForgetCardType(_In_ LPCTSTR szCardName)
  {
    const LONG nRet{__super::ForgetCardType(szCardName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void ForgetReader(_In_ LPCTSTR szReaderName)
  {
    const LONG nRet{__super::ForgetReader(szReaderName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void ForgetReaderGroup(_In_ LPCTSTR szGroupName)
  {
    const LONG nRet{__super::ForgetReaderGroup(szGroupName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void FreeMemory(_In_ LPCVOID pvMem)
  {
    const LONG nRet{__super::FreeMemory(pvMem)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void GetProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId,
                       _When_(_Old_(*pcchProvider) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)szProvider, _Outptr_result_buffer_all_(*pcchProvider)))
                       _When_(_Old_(*pcchProvider) != SCARD_AUTOALLOCATE, _Out_writes_to_(*pcchProvider, *pcchProvider) _Post_z_)
                       LPTSTR szProvider, _Inout_ LPDWORD pcchProvider)
  {
    const LONG nRet{__super::GetProviderName(szCardName, dwProviderId, szProvider, pcchProvider)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void GetProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId, _Out_ String& sProviderName)
  {
    const LONG nRet{__super::GetProviderName(szCardName, dwProviderId, sProviderName)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void GetDeviceTypeId(_In_ LPCTSTR szReaderName, _Inout_ LPDWORD pdwDeviceTypeId)
  {
    const LONG nRet{__super::GetDeviceTypeId(szReaderName, pdwDeviceTypeId)};
    CheckException(nRet);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#pragma warning(suppress: 26434)
  void GetProviderId(_In_ LPCTSTR szCard, _Out_ LPGUID pguidProviderId)
  {
    const LONG nRet{__super::GetProviderId(szCard, pguidProviderId)};
    CheckException(nRet);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void GetReaderDeviceInstanceId(_In_ LPCTSTR szReaderName,
                                 _When_(_Old_(*pcchDeviceInstanceId) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)szDeviceInstanceId, _Outptr_result_buffer_maybenull_(*pcchDeviceInstanceId) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                 _When_(_Old_(*pcchDeviceInstanceId) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchDeviceInstanceId) _Post_ _NullNull_terminated_)
                                 LPTSTR szDeviceInstanceId, _Inout_ LPDWORD pcchDeviceInstanceId)
  {
  #pragma warning(suppress: 6387)
    const LONG nRet{__super::GetReaderDeviceInstanceId(szReaderName, szDeviceInstanceId, pcchDeviceInstanceId)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void GetReaderDeviceInstanceId(_In_ LPCTSTR szReaderName, _Out_ String& sDeviceInstanceId)
  {
    const LONG nRet{__super::GetReaderDeviceInstanceId(szReaderName, sDeviceInstanceId)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void GetReaderIcon(_In_ LPCTSTR szReaderName,
                     _When_(_Old_(*pcbIcon) == SCARD_AUTOALLOCATE, _At_((LPBYTE*)pbIcon, _Outptr_result_bytebuffer_all_maybenull_(*pcbIcon)))
                     _When_(_Old_(*pcbIcon) != SCARD_AUTOALLOCATE, _Out_writes_bytes_to_(*pcbIcon, *pcbIcon) _Post_z_)
                     LPBYTE pbIcon, _Inout_ LPDWORD pcbIcon)
  {
    const LONG nRet{__super::GetReaderIcon(szReaderName, pbIcon, pcbIcon)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void GetReaderIcon(_In_ LPCTSTR szReaderName, _Out_ ByteArray& icon)
  {
    const LONG nRet{__super::GetReaderIcon(szReaderName, icon)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#pragma warning(suppress: 26434)
  void GetStatusChange(_In_ DWORD dwTimeout, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders)
  {
    const LONG nRet{__super::GetStatusChange(dwTimeout, rgReaderStates, cReaders)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void IntroduceCardType(_In_ LPCTSTR szCardName, _In_opt_  LPCGUID pguidPrimaryProvider, _In_opt_ LPCGUID rgguidInterfaces,
                         _In_ DWORD dwInterfaceCount, _In_ LPCBYTE pbAtr, _In_ LPCBYTE pbAtrMask, _In_ DWORD cbAtrLen)
  {
    const LONG nRet{__super::IntroduceCardType(szCardName, pguidPrimaryProvider, rgguidInterfaces, dwInterfaceCount, pbAtr, pbAtrMask, cbAtrLen)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void IntroduceReader(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szDeviceName)
  {
    const LONG nRet{__super::IntroduceReader(szReaderName, szDeviceName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void IntroduceReaderGroup(_In_ LPCTSTR szGroupName)
  {
    const LONG nRet{__super::IntroduceReaderGroup(szGroupName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void ListCards(_In_opt_ LPCBYTE pbAtr, _In_reads_opt_(cguidInterfaceCount) LPCGUID rgguidInterfaces, _In_ DWORD cguidInterfaceCount,
                 _When_(_Old_(*pcchCards) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszCards, _Outptr_result_buffer_maybenull_(*pcchCards) _At_(*_Curr_, _Post_ _NullNull_terminated_)))
                 _When_(_Old_(*pcchCards) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchCards) _Post_ _NullNull_terminated_)
                 LPTSTR mszCards, _Inout_ LPDWORD pcchCards)
  {
  #pragma warning(suppress: 6387)
    const LONG nRet{__super::ListCards(pbAtr, rgguidInterfaces, cguidInterfaceCount, mszCards, pcchCards)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListCards(_In_opt_ LPCBYTE pbAtr, _In_reads_opt_(cguidInterfaceCount) LPCGUID rgguidInterfaces, _In_ DWORD cguidInterfaceCount, _Out_ StringArray& cards)
  {
    const LONG nRet{__super::ListCards(pbAtr, rgguidInterfaces, cguidInterfaceCount, cards)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListInterfaces(_In_ LPCTSTR szCard, _Out_ LPGUID pguidInterfaces, _Inout_ LPDWORD pcguidInterfaces)
  {
    const LONG nRet{__super::ListInterfaces(szCard, pguidInterfaces, pcguidInterfaces)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListInterfaces(_In_ LPCTSTR szCard, _Out_ GUIDArray& interfaces)
  {
    const LONG nRet{__super::ListInterfaces(szCard, interfaces)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListReaderGroups(_Out_writes_opt_(*pcchGroups) _Post_ _NullNull_terminated_ LPTSTR mszGroups, _Inout_ LPDWORD pcchGroups)
  {
    const LONG nRet{__super::ListReaderGroups(mszGroups, pcchGroups)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void ListReaderGroups(_Out_ StringArray& groups)
  {
    const LONG nRet{__super::ListReaderGroups(groups)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListReaders(_In_opt_ LPCTSTR mszGroups,
                   _When_(_Old_(*pcchReaders) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaders, _Outptr_result_buffer_maybenull_(*pcchReaders) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                   _When_(_Old_(*pcchReaders) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaders) _Post_ _NullNull_terminated_)
                   LPTSTR mszReaders, _Inout_ LPDWORD pcchReaders)
  {
  #pragma warning(suppress: 6387)
    const LONG nRet{__super::ListReaders(mszGroups, mszReaders, pcchReaders)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void ListReaders(_In_opt_ LPCTSTR mszGroups, _Out_ StringArray& readers)
  {
    const LONG nRet{__super::ListReaders(mszGroups, readers)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void ListReadersWithDeviceInstanceId(_In_ LPCTSTR szDeviceInstanceId, 
                                       _When_(_Old_(*pcchReaders) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaders, _Outptr_result_buffer_maybenull_(*pcchReaders) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                       _When_(_Old_(*pcchReaders) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaders) _Post_ _NullNull_terminated_)
                                       LPTSTR mszReaders, _Inout_ LPDWORD pcchReaders)
  {
  #pragma warning(suppress: 6387)
    const LONG nRet{__super::ListReadersWithDeviceInstanceId(szDeviceInstanceId, mszReaders, pcchReaders)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#if (NTDDI_VERSION >= NTDDI_WIN8)
#pragma warning(suppress: 26434)
  void ListReadersWithDeviceInstanceId(_In_ LPCTSTR szDeviceInstanceId, _Out_ StringArray& readers)
  {
    const LONG nRet{__super::ListReadersWithDeviceInstanceId(szDeviceInstanceId, readers)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_WIN8)

#pragma warning(suppress: 26434)
  void LocateCards(_In_ LPCTSTR mszCards, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders)
  {
    const LONG nRet{__super::LocateCards(mszCards, rgReaderStates, cReaders)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void LocateCardsByATR(_In_ LPSCARD_ATRMASK rgAtrMasks, _In_ DWORD cAtrs, _Inout_ LPSCARD_READERSTATE rgReaderStates, _In_ DWORD cReaders)
  {
    const LONG nRet{__super::LocateCardsByATR(rgAtrMasks, cAtrs, rgReaderStates, cReaders)};
    CheckException(nRet);
  }

#if (NTDDI_VERSION >= NTDDI_VISTA)
#pragma warning(suppress: 26434)
  void ReadCache(_In_ UUID* pCardIdentifier, _In_ DWORD dwFreshnessCounter, _In_ LPTSTR szLookupName, _Out_writes_bytes_(*pdwDataLen) PBYTE pData, _Out_ DWORD* pdwDataLen)
  {
  #pragma warning(suppress: 6001)
    const LONG nRet{__super::ReadCache(pCardIdentifier, dwFreshnessCounter, szLookupName, pData, pdwDataLen)};
    CheckException(nRet);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

#if (NTDDI_VERSION >= NTDDI_VISTA)
#pragma warning(suppress: 26434)
  void WriteCache(_In_ UUID* pCardIdentifier, _In_ DWORD dwFreshnessCounter, _In_ LPTSTR szLookupName, _In_reads_bytes_(dwDataLen) PBYTE pData, _In_ DWORD dwDataLen)
  {
    const LONG nRet{__super::WriteCache(pCardIdentifier, dwFreshnessCounter, szLookupName, pData, dwDataLen)};
    CheckException(nRet);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

#pragma warning(suppress: 26434)
  void RemoveReaderFromGroup(_In_ LPCTSTR szReaderName, _In_ LPCTSTR szGroupName)
  {
    const LONG nRet{__super::RemoveReaderFromGroup(szReaderName, szGroupName)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void SetCardTypeProviderName(_In_ LPCTSTR szCardName, _In_ DWORD dwProviderId, _In_ LPCTSTR szProvider)
  {
    const LONG nRet{__super::SetCardTypeProviderName(szCardName, dwProviderId, szProvider)};
    CheckException(nRet);
  }
};


//Class which encapsulates an APDU request
class CSCARDWRAPPERS_EXT_CLASS CAPDURequest
{
public:
//Constructors / Destructors
  CAPDURequest() noexcept : m_byClass{0},
                            m_byIns{0},
                            m_byP1{0},
                            m_byP2{0},
                            m_byLE{0}
  {
  }

  CAPDURequest(_In_ BYTE byClass, _In_ BYTE byIns, _In_ BYTE byP1, _In_ BYTE byP2, _In_ const ByteArray& data, _In_ BYTE byLe = 0) : m_byClass{byClass},
                                                                                                                                     m_byIns{byIns},
                                                                                                                                     m_byP1{byP1},
                                                                                                                                     m_byP2{byP2},
                                                                                                                                     m_byLE{byLe}
  {
    m_Data = data;
  }

  CAPDURequest(_In_ BYTE byClass, _In_ BYTE byIns, _In_ BYTE byP1, _In_ BYTE byP2, _In_reads_bytes_(dwDataSize) const BYTE* pbyData = nullptr, _In_ DWORD dwDataSize = 0, _In_ BYTE byLe = 0) : m_byClass{byClass},
                                                                                                                                                                                                m_byIns{byIns},
                                                                                                                                                                                                m_byP1{byP1},
                                                                                                                                                                                                m_byP2{byP2},
                                                                                                                                                                                                m_byLE{byLe}
  {
#pragma warning(suppress: 26444 26481)
    m_Data.insert(m_Data.end(), pbyData, pbyData + dwDataSize);
  }

  CAPDURequest(const CAPDURequest&) = default;
  CAPDURequest(CAPDURequest&&) = default;
  ~CAPDURequest() = default;

//Methods
  CAPDURequest& operator=(const CAPDURequest&) = default;
  CAPDURequest& operator=(CAPDURequest&&) = default;

  bool Parse(_In_reads_bytes_(dwDataSize) const BYTE* pbyData, _In_ DWORD dwDataSize)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pbyData != nullptr);

    if (dwDataSize < 4)
      return false;

    DWORD dwCurrentOffset{0};
#pragma warning(suppress: 26481)
    m_byClass = pbyData[dwCurrentOffset++];
#pragma warning(suppress: 26481)
    m_byIns = pbyData[dwCurrentOffset++];
#pragma warning(suppress: 26481)
    m_byP1 = pbyData[dwCurrentOffset++];
#pragma warning(suppress: 26481)
    m_byP2 = pbyData[dwCurrentOffset++];

    m_byLE = 0;
    m_Data.resize(0);

    if (dwCurrentOffset < dwDataSize)
    {
#pragma warning(suppress: 26481)
      const DWORD dwPayloadSize{pbyData[dwCurrentOffset++]};
      if (dwDataSize < (dwCurrentOffset + dwPayloadSize))
        return false;

#pragma warning(suppress: 26481)
      const BYTE* pbyLocalData{&(pbyData[dwCurrentOffset])};
#pragma warning(suppress: 26444 26481)
      m_Data.insert(m_Data.end(), pbyLocalData, pbyLocalData + dwPayloadSize);
      dwCurrentOffset += dwPayloadSize;
    }

    if (dwCurrentOffset == (dwDataSize - 1))
#pragma warning(suppress: 26481)
      m_byLE = pbyData[dwCurrentOffset++];

    return (dwCurrentOffset == dwDataSize);
  }

  _Success_(return != 0) bool FormRawRequest(_Out_ ByteArray& request) const
  {
    //Get and validate the payload size
    const auto nPayloadSize{m_Data.size()};
    if (nPayloadSize > 0xFF)
      return false;
#pragma warning(suppress: 26472)
    auto dwPayloadSize{static_cast<DWORD>(nPayloadSize)};

    //Allocate and fill in the request buffer
    DWORD dwRequestSize{4 + dwPayloadSize};
    if (dwPayloadSize) //Include space for the Lc byte if required
      ++dwRequestSize;
    if (m_byLE) //Include space for the Le byte if required
      ++dwRequestSize;

    request.resize(dwRequestSize);

    //Fill in the common values
    int nRequestOffset{0};
#pragma warning(suppress: 26446)
    request[nRequestOffset++] = m_byClass;
#pragma warning(suppress: 26446)
    request[nRequestOffset++] = m_byIns;
#pragma warning(suppress: 26446)
    request[nRequestOffset++] = m_byP1;
#pragma warning(suppress: 26446)
    request[nRequestOffset++] = m_byP2;

    //Next need to do the payload data
    if (dwPayloadSize)
    {
#pragma warning(suppress: 26446 26472)
      request[nRequestOffset++] = static_cast<BYTE>(dwPayloadSize);
      for (DWORD i=0; i<dwPayloadSize; i++)
#pragma warning(suppress: 26446)
        request[nRequestOffset++] = m_Data[i];
    }

    if (m_byLE != 0)
#pragma warning(suppress: 26446 26489)
      request[nRequestOffset] = m_byLE;

    return true;
  }

//Member variables
  BYTE m_byClass; //The instruction class
  BYTE m_byIns; //The instruction code
  BYTE m_byP1; //The P1 parameter to the instruction
  BYTE m_byP2; //The P2 parameter to the instruction
  ByteArray m_Data; //Used defined data
  BYTE m_byLE; //The Length Expected field
};


//Class which encapsulates an APDU response
class CSCARDWRAPPERS_EXT_CLASS CAPDUResponse
{
public:
//Constructors / Destructors
  CAPDUResponse() noexcept : m_bySW1{0},
                             m_bySW2{0}
  {
  }

  CAPDUResponse(_In_ BYTE bySW1, _In_ BYTE bySW2, _In_reads_bytes_(dwDataSize) const BYTE* pbyData = nullptr, _In_ DWORD dwDataSize = 0) : m_bySW1{bySW1},
                                                                                                                                           m_bySW2{bySW2}
  {
#pragma warning(suppress: 26444 26481)
    m_Data.insert(m_Data.end(), pbyData, pbyData + dwDataSize);
  }

  CAPDUResponse(const CAPDUResponse&) = default;
  CAPDUResponse(CAPDUResponse&&) = default;
  ~CAPDUResponse() = default;

//Methods
  CAPDUResponse& operator=(const CAPDUResponse&) = default;
  CAPDUResponse& operator=(CAPDUResponse&&) = default;

#pragma warning(suppress: 26429)
  bool Parse(_In_reads_bytes_(dwDataSize) const BYTE* pbyData, _In_ DWORD dwDataSize)
  {
    if (dwDataSize < 2)
      return false;
    m_Data.clear();
#pragma warning(suppress: 26444 26481)
    m_Data.insert(m_Data.end(), pbyData, pbyData + dwDataSize - 2);

#pragma warning(suppress: 26481)
    m_bySW1 = pbyData[dwDataSize - 2];
#pragma warning(suppress: 26481)
    m_bySW2 = pbyData[dwDataSize - 1];

    return true;
  }

  _NODISCARD bool IsSuccessfulWithNoFurtherQualification() const noexcept
  {
    return (m_bySW1 == 0x90 && m_bySW2 == 0x00);
  }

  bool IsSuccessfulWithLEIndicated(BYTE& byLE) const noexcept
  {
    const bool bSuccess{m_bySW1 == 0x61};
    if (bSuccess)
      byLE = m_bySW2;
    return bSuccess;
  }
  
//Member variables
  BYTE m_bySW1; //The SW1 status bytes
  BYTE m_bySW2; //The SW2 status bytes
  ByteArray m_Data; //Used defined data
};


//Class which encapsulates a SCard card handle
class CSCARDWRAPPERS_EXT_CLASS CCard
{
public:
//Constructors / Destructors
#pragma warning(suppress: 26455)
  CCard(_In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD) noexcept : m_hCard{0},
                                                                          m_dwDisconnectDisposition{dwDisconnectDisposition},
                                                                          m_dwActiveProtocol{0},
                                                                          m_pCtx{nullptr}
  {
  }

  explicit CCard(_In_ SCARDHANDLE hCard, _In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD) noexcept : m_hCard{hCard},
                                                                                                           m_dwDisconnectDisposition{dwDisconnectDisposition},
                                                                                                           m_dwActiveProtocol{0},
                                                                                                           m_pCtx{nullptr}
  {
  }

  CCard(_In_ const CCard&) = delete;

#pragma warning(suppress: 26439 26495)
  CCard(_In_ CCard&& card) noexcept : m_hCard{0},
                                      m_dwDisconnectDisposition{card.m_dwDisconnectDisposition},
                                      m_dwActiveProtocol{card.m_dwActiveProtocol},
                                      m_pCtx{nullptr}
  {
    Attach(card.Detach());
  }

  ~CCard() noexcept
  {
    if (m_hCard != NULL)
      Disconnect(m_dwDisconnectDisposition);
  }

//Methods
  CCard& operator=(_In_ const CCard&) = delete;

  CCard& operator=(_In_ CCard&& card) noexcept
  {
    if (m_hCard != NULL)
      Disconnect();
    Attach(card.Detach());
    return *this;
  }

  LONG Connect(_In_ CContext& ctx, _In_ LPCTSTR szReader, _In_ DWORD dwShareMode = SCARD_SHARE_SHARED, _In_ DWORD dwPreferredProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard == NULL);

    //Hive away the context pointer
    m_pCtx = &ctx;

    return SCardConnect(ctx, szReader, dwShareMode, dwPreferredProtocols, &m_hCard, &m_dwActiveProtocol);
  }

  _NODISCARD DWORD GetActiveProtocol() const noexcept
  {
    return m_dwActiveProtocol;
  }

  _NODISCARD DWORD GetDisconnectDisposition() const noexcept
  {
    return m_dwDisconnectDisposition;
  }

  LONG Disconnect(_In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD) noexcept
  {
    LONG nRet{SCARD_S_SUCCESS};
    if (m_hCard != NULL)
    {
      nRet = SCardDisconnect(m_hCard, dwDisconnectDisposition);
      m_hCard = NULL;
      m_pCtx = nullptr;
    }
    return nRet;
  }

  LONG BeginTransaction() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardBeginTransaction(m_hCard);
  }

  LONG EndTransaction(_In_ DWORD dwDisposition = SCARD_LEAVE_CARD) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardEndTransaction(m_hCard, dwDisposition);
  }

  LONG Control(_In_ DWORD dwControlCode, _In_reads_bytes_(cbInBufferSize) LPCVOID lpInBuffer, _In_ DWORD cbInBufferSize,
               _Out_writes_bytes_(cbOutBufferSize) LPVOID lpOutBuffer, _In_ DWORD cbOutBufferSize, _Out_ LPDWORD lpBytesReturned) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardControl(m_hCard, dwControlCode, lpInBuffer, cbInBufferSize, lpOutBuffer, cbOutBufferSize, lpBytesReturned);
  }

  operator SCARDHANDLE() const noexcept
  {
    return m_hCard;
  }

  void Attach(_In_ SCARDHANDLE hCard) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard == NULL);

    m_hCard = hCard;
  }

  SCARDHANDLE Detach() noexcept
  {
    const SCARDHANDLE h{m_hCard};
    m_hCard = NULL;
    return h;
  }

  LONG GetAttrib(_In_ DWORD dwAttrID, _Out_writes_bytes_opt_(*pcbAttrLen) LPBYTE pbAttr, _Inout_ LPDWORD pcbAttrLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardGetAttrib(m_hCard, dwAttrID, pbAttr, pcbAttrLen);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG GetAttrib(_In_ DWORD dwAttrID, _Out_ ByteArray& Attr)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(m_pCtx != nullptr);

    //First get the length of the data
    DWORD dwAttrLen{SCARD_AUTOALLOCATE};
    BYTE* pbyAttr{nullptr};
#pragma warning(suppress: 26490)
    const LONG nRet{GetAttrib(dwAttrID, reinterpret_cast<LPBYTE>(&pbyAttr), &dwAttrLen)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;

    Attr.clear();
#pragma warning(suppress: 26444 26481 26489)
    Attr.insert(Attr.end(), pbyAttr, pbyAttr + dwAttrLen);
    m_pCtx->FreeMemory(pbyAttr);

    return nRet;
  }
 
  _Success_(return == SCARD_S_SUCCESS) LONG GetAttrib(_In_ DWORD dwAttrID, _Out_ DWORD& dwAttr)
  {
    ByteArray attr;
    LONG nRet{GetAttrib(dwAttrID, attr)};
    if (nRet == SCARD_S_SUCCESS)
    {
      if (attr.size() == sizeof(DWORD))
      {
#pragma warning(suppress: 26429 26490)
        auto pdwValue{reinterpret_cast<const DWORD*>(attr.data())};
        dwAttr = *pdwValue;
      }
      else
        nRet = SCARD_E_INVALID_VALUE;
    }
    return nRet;
  }

  _Success_(return == SCARD_S_SUCCESS) LONG GetAttrib(_In_ DWORD dwAttrID, _Out_ BYTE& byAttr)
  {
    ByteArray attr;
    LONG nRet{GetAttrib(dwAttrID, attr)};
    if (nRet == SCARD_S_SUCCESS)
    {
      if (attr.size() == sizeof(BYTE))
#pragma warning(suppress: 26446)
        byAttr = attr[0];
      else
        nRet = SCARD_E_INVALID_VALUE;
    }
    return nRet;
  }

#if (NTDDI_VERSION >= NTDDI_VISTA)
  _Success_(return == SCARD_S_SUCCESS) LONG GetTransmitCount(_Out_ LPDWORD pcTransmitCount) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardGetTransmitCount(m_hCard, pcTransmitCount);
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

  LONG Reconnect(_In_ DWORD dwShareMode = SCARD_SHARE_SHARED, _In_ DWORD dwPreferredProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, _In_ DWORD dwInitialization = SCARD_LEAVE_CARD) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardReconnect(m_hCard, dwShareMode, dwPreferredProtocols, dwInitialization, &m_dwActiveProtocol);
  }

  LONG SetAttrib(_In_ DWORD dwAttrId, _In_reads_bytes_(cbAttrLen) LPCBYTE pbAttr, _In_ DWORD cbAttrLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardSetAttrib(m_hCard, dwAttrId, pbAttr, cbAttrLen);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG Status(_When_(_Old_(*pcchReaderLen) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaderNames, _Outptr_result_buffer_maybenull_(*pcchReaderLen) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
                                                   _When_(_Old_(*pcchReaderLen) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaderLen) _Post_ _NullNull_terminated_)
                                                   LPTSTR mszReaderNames, _Inout_opt_ LPDWORD pcchReaderLen, _Out_opt_ LPDWORD pdwState, _Out_opt_ LPDWORD pdwProtocol,
                                                   _When_(_Old_(*pcbAtrLen) == SCARD_AUTOALLOCATE, _At_((LPBYTE*)pbAtr, _Outptr_result_buffer_maybenull_(*pcbAtrLen) _At_(*_Curr_, _Post_ _NullNull_terminated_)))
                                                   _When_(_Old_(*pcbAtrLen) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcbAtrLen) _Post_ _NullNull_terminated_)
                                                   LPBYTE pbAtr, _Inout_opt_ LPDWORD pcbAtrLen) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

  #pragma warning(suppress: 6387)
    return SCardStatus(m_hCard, mszReaderNames, pcchReaderLen, pdwState, pdwProtocol, pbAtr, pcbAtrLen);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG Status(_Out_ StringArray& readerName, _Out_opt_ LPDWORD pdwState, _Out_opt_ LPDWORD pdwProtocol, _Out_ ByteArray& Atr)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);
#pragma warning(suppress: 26477)
    ATLASSUME(m_pCtx != nullptr);

    Atr.resize(32);
#pragma warning(suppress: 26472)
    auto dwAtrLen{static_cast<DWORD>(Atr.size())};
    LPTSTR szReaderName{nullptr};
    auto dwReaderLen{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
    const LONG nRet{Status(reinterpret_cast<LPTSTR>(&szReaderName), &dwReaderLen, pdwState, pdwProtocol, Atr.data(), &dwAtrLen)};
    if (nRet != SCARD_S_SUCCESS)
      return nRet;
    MultiSZToStringArray(szReaderName, readerName);
#pragma warning(suppress: 26489)
    Atr.resize(dwAtrLen);
    if (szReaderName != nullptr)
      m_pCtx->FreeMemory(szReaderName);

    return nRet;
  }

  _Success_(return == SCARD_S_SUCCESS) LONG Transmit(_In_ LPCSCARD_IO_REQUEST pioSendPci, _In_reads_bytes_(cbSendLength) LPCBYTE pbSendBuffer, _In_ DWORD cbSendLength, _Inout_opt_ LPSCARD_IO_REQUEST pioRecvPci,
                                                     _Out_writes_bytes_(*pcbRecvLength) LPBYTE pbRecvBuffer, _Inout_ LPDWORD pcbRecvLength) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCard != NULL);

    return SCardTransmit(m_hCard, pioSendPci, pbSendBuffer, cbSendLength, pioRecvPci, pbRecvBuffer, pcbRecvLength);
  }

  _Success_(return == SCARD_S_SUCCESS) LONG Transmit(_In_ const CAPDURequest& request, _Out_ CAPDUResponse& response)
  {
    //Form the raw request
    ByteArray rawRequest;
    if (!request.FormRawRequest(rawRequest))
      return SCARD_E_INVALID_VALUE;
#pragma warning(suppress: 26472)
    const auto dwRequestSize{static_cast<DWORD>(rawRequest.size())};

    //Allocate the raw response buffer
#pragma warning(suppress: 26472)
    auto dwResponseSize{static_cast<DWORD>(2 + request.m_byLE)};
    ByteArray responseBuffer{dwResponseSize, std::allocator<BYTE>{}};
#pragma warning(suppress: 26446)
    BYTE* lpResponse{responseBuffer.data()};

    //Call the other version of the Transmit method
    LONG nRet{SCARD_E_PROTO_MISMATCH};
    switch (m_dwActiveProtocol)
    {
      case SCARD_PROTOCOL_T0:
      {
        nRet = Transmit(SCARD_PCI_T0, rawRequest.data(), dwRequestSize, nullptr, lpResponse, &dwResponseSize);
        break;
      }
      case SCARD_PROTOCOL_T1:
      {
        nRet = Transmit(SCARD_PCI_T1, rawRequest.data(), dwRequestSize, nullptr, lpResponse, &dwResponseSize);
        break;
      }
      case SCARD_PROTOCOL_RAW:
      {
        nRet = Transmit(SCARD_PCI_RAW, rawRequest.data(), dwRequestSize, nullptr, lpResponse, &dwResponseSize);
        break;
      }
      default:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(false);
        break;
      }
    }

    //parse the raw buffer into the response object
    if (nRet == SCARD_S_SUCCESS)
    {
      if (!response.Parse(lpResponse, dwResponseSize))
        nRet = SCARD_E_INVALID_VALUE;
    }

    return nRet;
  }

protected:
//Member variables
  SCARDHANDLE m_hCard;
  DWORD m_dwDisconnectDisposition;
  DWORD m_dwActiveProtocol;
  CContext* m_pCtx; //The context from which this card was created
};


//exception based version of CCard
class CSCARDWRAPPERS_EXT_CLASS CCard2 : public CCard
{
public:
//Constructors / Destructors
#pragma warning(suppress: 26455)
  CCard2(_In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD) noexcept : CCard{}
  {
    m_dwDisconnectDisposition = dwDisconnectDisposition;
  }

  explicit CCard2(_In_ SCARDHANDLE hCard, _In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD) noexcept : CCard{hCard, dwDisconnectDisposition}
  {
  }

  CCard2(_In_ const CCard2&) = delete;

#pragma warning(suppress: 26439)
  CCard2(_In_ CCard2&& card) noexcept : CCard{std::move(card)}
  {
  }

  ~CCard2() = default;

//Methods
  CCard2& operator=(_In_ const CCard2&) = delete;

#pragma warning(suppress: 26456)
  CCard2& operator=(_In_ CCard2&& card) noexcept
  {
    __super::operator=(std::move(card));
    return *this;
  }

#pragma warning(suppress: 26434)
  void Connect(_In_ CContext2& ctx, _In_ LPCTSTR szReader, _In_ DWORD dwShareMode = SCARD_SHARE_SHARED, _In_ DWORD dwPreferredProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1)
  {
    const LONG nRet{__super::Connect(ctx, szReader, dwShareMode, dwPreferredProtocols)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void Disconnect(_In_ DWORD dwDisconnectDisposition = SCARD_LEAVE_CARD)
  {
    const LONG nRet{__super::Disconnect(dwDisconnectDisposition)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void BeginTransaction()
  {
    const LONG nRet{__super::BeginTransaction()};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void EndTransaction(_In_ DWORD dwDisposition = SCARD_LEAVE_CARD)
  {
    const LONG nRet{__super::EndTransaction(dwDisposition)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void Control(_In_ DWORD dwControlCode, _In_reads_bytes_(cbInBufferSize) LPCVOID lpInBuffer, _In_ DWORD cbInBufferSize, 
               _Out_writes_bytes_(cbOutBufferSize) LPVOID lpOutBuffer, _In_ DWORD cbOutBufferSize, _Out_ LPDWORD lpBytesReturned)
  {
    const LONG nRet{__super::Control(dwControlCode, lpInBuffer, cbInBufferSize, lpOutBuffer, cbOutBufferSize, lpBytesReturned)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void GetAttrib(_In_ DWORD dwAttrID, _Out_writes_bytes_opt_(*pcbAttrLen) LPBYTE pbAttr, _Inout_ LPDWORD pcbAttrLen)
  {
    const LONG nRet{__super::GetAttrib(dwAttrID, pbAttr, pcbAttrLen)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void GetAttrib(_In_ DWORD dwAttrID, _Out_ ByteArray& Attr)
  {
    const LONG nRet{__super::GetAttrib(dwAttrID, Attr)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
 
  DWORD GetAttribDWORD(_In_ DWORD dwAttrID)
  {
    DWORD dwAttr{0};
    const LONG nRet{__super::GetAttrib(dwAttrID, dwAttr)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
    return dwAttr;
  }

  BYTE GetAttribBYTE(_In_ DWORD dwAttrID)
  {
    BYTE byAttr{0};
    const LONG nRet{__super::GetAttrib(dwAttrID, byAttr)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
    return byAttr;
  }

#if (NTDDI_VERSION >= NTDDI_VISTA)
#pragma warning(suppress: 26434)
  DWORD GetTransmitCount()
  {
    DWORD dwTransmitCount{0};
    const LONG nRet{__super::GetTransmitCount(&dwTransmitCount)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
    return dwTransmitCount;
  }
#endif //#if (NTDDI_VERSION >= NTDDI_VISTA)

#pragma warning(suppress: 26434)
  void Reconnect(_In_ DWORD dwShareMode = SCARD_SHARE_SHARED, _In_ DWORD dwPreferredProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, _In_ DWORD dwInitialization = SCARD_LEAVE_CARD)
  {
    const LONG nRet{__super::Reconnect(dwShareMode, dwPreferredProtocols, dwInitialization)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void SetAttrib(_In_ DWORD dwAttrId, _In_reads_bytes_(cbAttrLen) LPCBYTE pbAttr, _In_ DWORD cbAttrLen)
  {
    const LONG nRet{__super::SetAttrib(dwAttrId, pbAttr, cbAttrLen)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void Status(_When_(_Old_(*pcchReaderLen) == SCARD_AUTOALLOCATE, _At_((LPTSTR*)mszReaderNames, _Outptr_result_buffer_maybenull_(*pcchReaderLen) _At_(*_Curr_, _Post_z_ _Post_ _NullNull_terminated_)))
              _When_(_Old_(*pcchReaderLen) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcchReaderLen) _Post_ _NullNull_terminated_)
              LPTSTR mszReaderNames, _Inout_opt_ LPDWORD pcchReaderLen, _Out_opt_ LPDWORD pdwState, _Out_opt_ LPDWORD pdwProtocol,
              _When_(_Old_(*pcbAtrLen) == SCARD_AUTOALLOCATE, _At_((LPBYTE*)pbAtr, _Outptr_result_buffer_maybenull_(*pcbAtrLen) _At_(*_Curr_, _Post_ _NullNull_terminated_)))
              _When_(_Old_(*pcbAtrLen) != SCARD_AUTOALLOCATE, _Out_writes_opt_(*pcbAtrLen) _Post_ _NullNull_terminated_)
              LPBYTE pbAtr, _Inout_opt_ LPDWORD pcbAtrLen)
  {
  #pragma warning(suppress: 6387)
    const LONG nRet{__super::Status(mszReaderNames, pcchReaderLen, pdwState, pdwProtocol, pbAtr, pcbAtrLen)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void Status(_Out_ StringArray& readerName, _Out_opt_ LPDWORD pdwState, _Out_opt_ LPDWORD pdwProtocol, _Out_ ByteArray& Atr)
  {
    const LONG nRet{__super::Status(readerName, pdwState, pdwProtocol, Atr)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }

#pragma warning(suppress: 26434)
  void Transmit(_In_ LPCSCARD_IO_REQUEST pioSendPci, _In_reads_bytes_(cbSendLength) LPCBYTE pbSendBuffer, _In_ DWORD cbSendLength, _Inout_opt_ LPSCARD_IO_REQUEST pioRecvPci,
                _Out_writes_bytes_(*pcbRecvLength) LPBYTE pbRecvBuffer, _Inout_ LPDWORD pcbRecvLength)
  {
    const LONG nRet{__super::Transmit(pioSendPci, pbSendBuffer, cbSendLength, pioRecvPci, pbRecvBuffer, pcbRecvLength)};
    CheckException(nRet);
  }

#pragma warning(suppress: 26434)
  void Transmit(_In_ const CAPDURequest& request, _Out_ CAPDUResponse& response)
  {
    const LONG nRet{__super::Transmit(request, response)};
    CheckException(nRet);
#pragma warning(suppress: 26477)
    ATLASSUME(nRet == SCARD_S_SUCCESS);
  }
};


//Class which implements an ATR parser
class CATR
{
public:
//Enums
  enum class ParseError
  {
    NoError = 0,
    TSNotFound = 1,
    UnknownTSValue = 2,
    T0NotFound = 3,
    TA1NotFound = 4,
    TA2NotFound = 5,
    TA3NotFound = 6,
    TA4NotFound = 7,
    TB1NotFound = 8,
    TB2NotFound = 9,
    TB3NotFound = 10,
    TB4NotFound = 11,
    TC1NotFound = 12,
    TC2NotFound = 13,
    TC3NotFound = 14,
    TC4NotFound = 15,
    TD1NotFound = 16,
    TD2NotFound = 17,
    TD3NotFound = 18,
    TD4NotFound = 19,
    HistoricalCharactersNotFound = 20,
    TCKIsNotLast = 21
  };

//Constructors / Destructors
  CATR() noexcept : m_bTSPresent{false},
                    m_byTS{0x00},
                    m_bDirectConvention{false},
                    m_byT0{0x00},
                    m_nY1{-1},
                    m_nK{-1},
                    m_bTA1Present{false},
                    m_byTA1{0x00},
                    m_nFI{-1},
                    m_nF{-1},
                    m_fMax{-1},
                    m_nDI{-1},
                    m_nD{-1},
                    m_bTB1Present{false},
                    m_byTB1{0x00},
                    m_nPI1{-1},
                    m_nII{-1},
                    m_bTC1Present{false},
                    m_byTC1{0x00},
                    m_nN{-1},
                    m_bTD1Present{false},
                    m_byTD1{0x00},
                    m_nY2{-1},
                    m_nProtocol1{-1},
                    m_bTA2Present{false},
                    m_byTA2{0x00},
                    m_bSwitchBetweenNegotiableModeAndSpecificModePossible{false},
                    m_bTransmissionCharactersImplicitlyDefined{false},
                    m_nTA2Protocol{-1},
                    m_bTB2Present{false},
                    m_byTB2{0x00},
                    m_nPI2{-1},
                    m_bTC2Present{false},
                    m_byTC2{0x00},
                    m_nWI{-1},
                    m_bTD2Present{false},
                    m_byTD2{0x00},
                    m_nY3{-1},
                    m_nProtocol2{-1},
                    m_bTA3Present{false},
                    m_byTA3{0x00},
                    m_bTB3Present{false},
                    m_byTB3{0x00},
                    m_nTB3CWI{-1},
                    m_nTB3BWI{-1},
                    m_bTC3Present{false},
                    m_byTC3{0x00},
                    m_bTC3CRCUsed{false},
                    m_bTD3Present{false},
                    m_byTD3{0x00},
                    m_nY4{-1},
                    m_nProtocol3{-1},
                    m_bTA4Present{false},
                    m_byTA4{0x00},
                    m_bTB4Present{false},
                    m_byTB4{0x00},
                    m_nTB4CWI{-1},
                    m_nTB4BWI{-1},
                    m_bTC4Present{false},
                    m_byTC4{0x00},
                    m_bTC4CRCUsed{false},
                    m_bTD4Present{false},
                    m_byTD4{0x00},
                    m_nY5{-1},
                    m_nProtocol4{-1},
                    m_bTCKPresent{false},
                    m_byTCK{0x00},
                    m_bTCKValid{false}
  {
  }

//Member variables
  bool m_bTSPresent;
  BYTE m_byTS;
  bool m_bDirectConvention;
  BYTE m_byT0;
  int m_nY1;
  int m_nK;
  bool m_bTA1Present;
  BYTE m_byTA1;
  int m_nFI;
  int m_nF;
  double m_fMax;
  int m_nDI;
  int m_nD;
  bool m_bTB1Present;
  BYTE m_byTB1;
  int m_nPI1;
  int m_nII;
  bool m_bTC1Present;
  BYTE m_byTC1;
  int m_nN;
  bool m_bTD1Present;
  BYTE m_byTD1;
  int m_nY2;
  int m_nProtocol1;
  bool m_bTA2Present;
  BYTE m_byTA2;
  bool m_bSwitchBetweenNegotiableModeAndSpecificModePossible;
  bool m_bTransmissionCharactersImplicitlyDefined;
  int m_nTA2Protocol;
  bool m_bTB2Present;
  BYTE m_byTB2;
  int m_nPI2;
  bool m_bTC2Present;
  BYTE m_byTC2;
  int m_nWI;
  bool m_bTD2Present;
  BYTE m_byTD2;
  int m_nY3;
  int m_nProtocol2;
  bool m_bTA3Present;
  BYTE m_byTA3;
  bool m_bTB3Present;
  BYTE m_byTB3;
  int m_nTB3CWI;
  int m_nTB3BWI;
  bool m_bTC3Present;
  BYTE m_byTC3;
  bool m_bTC3CRCUsed;
  bool m_bTD3Present;
  BYTE m_byTD3;
  int m_nY4;
  int m_nProtocol3;
  bool m_bTA4Present;
  BYTE m_byTA4;
  bool m_bTB4Present;
  BYTE m_byTB4;
  int m_nTB4CWI;
  int m_nTB4BWI;
  bool m_bTC4Present;
  BYTE m_byTC4;
  bool m_bTC4CRCUsed;
  bool m_bTD4Present;
  BYTE m_byTD4;
  int m_nY5;
  int m_nProtocol4;
  bool m_bTCKPresent;
  BYTE m_byTCK;
  bool m_bTCKValid;
  ByteArray m_HistoricalCharacters;

//Methods
  _Success_(return == true) bool Parse(_In_ SCard::ByteArray& ATR, _Out_ ParseError& error)
  {
    //Reset the data
    *this = CATR();

    //Set the output parameter to a default value
    error = ParseError::NoError;

#pragma warning(suppress: 26472)
    const int nSize{static_cast<int>(ATR.size())};
    int nCurrentIndex{0};

    //TS details
    if (!ParseTS(ATR, nSize, nCurrentIndex, error))
      return false;

    //T0 details
    if (!ParseT0(ATR, nSize, nCurrentIndex, error))
      return false;

    //TA1 details
    if (m_bTA1Present)
    {
      if (nCurrentIndex < nSize)
#pragma warning(suppress: 26486 26489)
        ParseTA1(ATR, nCurrentIndex);
      else
      {
        error = ParseError::TA1NotFound;
        return false;
      }
    }
    
    //TB1 details
    if (m_bTB1Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTB1Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTB1 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nPI1 = static_cast<int>(m_byTB1 & 0x1F);
#pragma warning(suppress: 26472)
        m_nII = static_cast<int>((m_byTB1 & 0x60) >> 5);
      }
      else
      {
        error = ParseError::TB1NotFound;
        return false;
      }
    }

    //TC1 details
    if (m_bTC1Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTC1Present = true;
#pragma warning(suppress: 26489)
#pragma warning(suppress: 26446)
        m_byTC1 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nN =  static_cast<int>(m_byTC1);
      }
      else
      {
        error = ParseError::TC1NotFound;
        return false;
      }
    }

    //TD1 details
    if (m_bTD1Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTD1Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTD1 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nY2 = static_cast<int>((m_byTD1 & 0xF0) >> 4);
#pragma warning(suppress: 26472)
        m_nProtocol1 = static_cast<int>(m_byTD1 & 0x0F);
        if (m_byTD1 & 0x10)
          m_bTA2Present = true;
        if (m_byTD1 & 0x20)
          m_bTB2Present = true;
        if (m_byTD1 & 0x40)
           m_bTC2Present = true;
        if (m_byTD1 & 0x80)
          m_bTD2Present = true;
      }
      else
      {
        error = ParseError::TD1NotFound;
        return false;
      }
    }

    //TA2 details
    if (m_bTA2Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTA2Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTA2 = ATR[nCurrentIndex++];
        if (m_byTA2 & 0x80)
          m_bSwitchBetweenNegotiableModeAndSpecificModePossible = true;
        if (m_byTA2 & 0x10)
          m_bTransmissionCharactersImplicitlyDefined = true;
#pragma warning(suppress: 26472)
        m_nTA2Protocol = static_cast<int>(m_byTA2 & 0x0F);
      }
      else
      {
        error = ParseError::TA2NotFound;
        return false;
      }
    }

    //TB2 details
    if (m_bTB2Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTB2Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTB2 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nPI2 = static_cast<int>(m_byTB2 & 0x1F);
      }
      else
      {
        error = ParseError::TB2NotFound;
        return false;
      }
    }

    //TC2 details
    if (m_bTC2Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTC2Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTC2 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nWI = static_cast<int>(m_byTC2);
      }
      else
      {
        error = ParseError::TC2NotFound;
        return false;
      }
    }

    //TD2 details
    if (m_bTD2Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTD2Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTD2 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nY3 = static_cast<int>((m_byTD2 & 0xF0) >> 4);
#pragma warning(suppress: 26472)
        m_nProtocol2 = static_cast<int>(m_byTD2 & 0x0F);
        if (m_byTD2 & 0x10)
          m_bTA3Present = true;
        if (m_byTD2 & 0x20)
          m_bTB3Present = true;
        if (m_byTD2 & 0x40)
           m_bTC3Present = true;
        if (m_byTD2 & 0x80)
          m_bTD3Present = true;
      }
      else
      {
        error = ParseError::TD2NotFound;
        return false;
      }
    }

    //TA3 details
    if (m_bTA3Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTA3Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTA3 = ATR[nCurrentIndex++];
      }
      else
      {
        error = ParseError::TA3NotFound;
        return false;
      }
    }

    //TB3 details
    if (m_bTB3Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTB3Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTB3 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nTB3CWI = static_cast<int>((m_byTB3 & 0x0F));
#pragma warning(suppress: 26472)
        m_nTB3BWI = static_cast<int>((m_byTB3 & 0xF0) >> 4);
      }
      else
      {
        error = ParseError::TB3NotFound;
        return false;
      }
    }

    //TC3 details
    if (m_bTC3Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTC3Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTC3 = ATR[nCurrentIndex++];
        if (m_byTC3 & 0x01)
          m_bTC3CRCUsed = true;
      }
      else
      {
        error = ParseError::TC3NotFound;
        return false;
      }
    }

    //TD3 details
    if (m_bTD3Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTD3Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTD3 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nY4 = static_cast<int>((m_byTD3 & 0xF0) >> 4);
#pragma warning(suppress: 26472)
        m_nProtocol3 = static_cast<int>(m_byTD3 & 0x0F);
        if (m_byTD3 & 0x10)
          m_bTA4Present = true;
        if (m_byTD3 & 0x20)
          m_bTB4Present = true;
        if (m_byTD3 & 0x40)
           m_bTC4Present = true;
        if (m_byTD3 & 0x80)
          m_bTD4Present = true;
      }
      else
      {
        error = ParseError::TD3NotFound;
        return false;
      }
    }

    //TA4 details
    if (m_bTA4Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTA4Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTA4 = ATR[nCurrentIndex++];
      }
      else
      {
        error = ParseError::TA4NotFound;
        return false;
      }
    }

    //TB4 details
    if (m_bTB4Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTB4Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTB4 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nTB4CWI = static_cast<int>((m_byTB4 & 0x0F));
#pragma warning(suppress: 26472)
        m_nTB4BWI = static_cast<int>((m_byTB4 & 0xF0) >> 4);
      }
      else
      {
        error = ParseError::TB4NotFound;
        return false;
      }
    }

    //TC4 details
    if (m_bTC4Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTC4Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTC4 = ATR[nCurrentIndex++];
        if (m_byTC4 & 0x01)
          m_bTC4CRCUsed = true;
      }
      else
      {
        error = ParseError::TC4NotFound;
        return false;
      }
    }

    //TD4 details
    if (m_bTD4Present)
    {
      if (nCurrentIndex < nSize)
      {
        m_bTD4Present = true;
#pragma warning(suppress: 26446 26489)
        m_byTD4 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
        m_nY5 = static_cast<int>((m_byTD4 & 0xF0) >> 4);
#pragma warning(suppress: 26472)
        m_nProtocol4 = static_cast<int>(m_byTD4 & 0x0F);
      }
      else
      {
        error = ParseError::TD4NotFound;
        return false;
      }
    }

    //T1-TK details
    if (m_nK)
    {
      if ((nCurrentIndex + m_nK) <= nSize)
      {
#pragma warning(suppress: 26446 26489)
        BYTE* pbyHistoricalCharacters{&(ATR[nCurrentIndex])};
        m_HistoricalCharacters.clear();
#pragma warning(suppress: 26444 26481)
        m_HistoricalCharacters.insert(m_HistoricalCharacters.end(), pbyHistoricalCharacters, pbyHistoricalCharacters + m_nK);
        nCurrentIndex += m_nK;
      }
      else
      {
        error = ParseError::HistoricalCharactersNotFound;
        return false;
      }
    }

    //TCK details
    if (nCurrentIndex == (nSize - 1))
    {
      m_bTCKPresent = true;

      BYTE byCalculatedChecksum{0};
      for (int i=1; i<(nSize-1); i++)
#pragma warning(suppress: 26446 26489)
        byCalculatedChecksum ^= ATR[i];
#pragma warning(suppress: 26446 26472 26489)
      m_byTCK = static_cast<int>(ATR[nCurrentIndex]);
#pragma warning(suppress: 26446 26489)
      m_bTCKValid = (byCalculatedChecksum == ATR[nCurrentIndex]);
    }
    else if (nCurrentIndex < nSize)
    {
      error = ParseError::TCKIsNotLast;
      return false;
    }

    return true;
  }

protected:
#pragma warning(suppress: 26440)
  _Success_(return == true) bool ParseTS(_In_ const SCard::ByteArray& ATR, _In_ int nSize, _Inout_ int& nCurrentIndex, _Out_ ParseError& error)
  {
    //Set the output parameter to a default value
    error = ParseError::NoError;

    if (nCurrentIndex < nSize)
    {
#pragma warning(suppress: 26446)
      m_byTS = ATR[nCurrentIndex++];
      if (m_byTS == 0x3F)
        m_bDirectConvention = false;
      else if (m_byTS == 0x3B)
        m_bDirectConvention = true;
      else
      {
        error = ParseError::UnknownTSValue;
        return false;
      }
    }
    else
    {
      error = ParseError::TSNotFound;
      return false;
    }

    return true;
  }

#pragma warning(suppress: 26440)
  _Success_(return == true) bool ParseT0(_In_ const SCard::ByteArray& ATR, _In_ int nSize, _Inout_ int& nCurrentIndex, _Out_ ParseError& error)
  {
    //Set the output parameter to a default value
    error = ParseError::NoError;

    if (nCurrentIndex < nSize)
    {
#pragma warning(suppress: 26446)
      m_byT0 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
      m_nY1 = static_cast<int>((m_byT0 & 0xF0) >> 4);
#pragma warning(suppress: 26472)
      m_nK = static_cast<int>(m_byT0 & 0x0F);

      if (m_byT0 & 0x10)
        m_bTA1Present = true;
      if (m_byT0 & 0x20)
        m_bTB1Present = true;
      if (m_byT0 & 0x40)
         m_bTC1Present = true;
      if (m_byT0 & 0x80)
        m_bTD1Present = true;
    }
    else
    {
      error = ParseError::T0NotFound;
      return false;
    }

    return true;
  }

#pragma warning(suppress: 26440)
  void ParseTA1(_In_ const SCard::ByteArray& ATR, _Inout_ int& nCurrentIndex)
  {
    m_bTA1Present = true;
#pragma warning(suppress: 26446)
    m_byTA1 = ATR[nCurrentIndex++];
#pragma warning(suppress: 26472)
    m_nFI = static_cast<int>((m_byTA1 & 0xF0) >> 4);
    switch (m_nFI)
    {
      case 0:
      {
        m_nF = 372;
        m_fMax = 4;
        break;
      }
      case 1:
      {
        m_nF = 372;
        m_fMax = 5;
        break;
      }
      case 2:
      {
        m_nF = 558;
        m_fMax = 6;
        break;
      }
      case 3:
      {
        m_nF = 744;
        m_fMax = 8;
        break;
      }
      case 4:
      {
        m_nF = 1116;
        m_fMax = 12;
        break;
      }
      case 5:
      {
        m_nF = 1488;
        m_fMax = 16;
        break;
      }
      case 6:
      {
        m_nF = 1860;
        m_fMax = 20;
        break;
      }
      case 17:
      {
        m_nF = 512;
        m_fMax = 5;
        break;
      }
      case 18:
      {
        m_nF = 768;
        m_fMax = 7.5;
        break;
      }
      case 19:
      {
        m_nF = 1024;
        m_fMax = 10;
        break;
      }
      case 20:
      {
        m_nF = 1536;
        m_fMax = 15;
        break;
      }
      case 21:
      {
        m_nF = 2048;
        m_fMax = 20;
        break;
      }
      default:
      {
        break;
      }
    }
#pragma warning(suppress: 26472)
    m_nDI = static_cast<int>(m_byTA1 & 0x0F);
    switch (m_nDI)
    {
      case 1:
      {
        m_nD = 1;
        break;
      }
      case 2:
      {
        m_nD = 2;
        break;
      }
      case 3:
      {
        m_nD = 4;
        break;
      }
      case 4:
      {
        m_nD = 8;
        break;
      }
      case 5:
      {
        m_nD = 16;
        break;
      }
      case 6:
      {
        m_nD = 32;
        break;
      }
      case 8:
      {
        m_nD = 12;
        break;
      }
      case 9:
      {
        m_nD = 20;
        break;
      }
      default:
      {
        break;
      }
    }
  }
};


}; //namespace SCard


#endif //__SCARDWRAPPERS_H__
